import unittest
import gettextpo

class PoFileTestCase(unittest.TestCase):
    def testCreateEmpty(self):
        '''Test that we can create an empty pofile object'''
        pofile = gettextpo.PoFile()
        self.assertEquals(list(iter(pofile)), [])
    def testAddMessage(self):
        '''Test that we can add messages to a new pofile object'''
        pofile = gettextpo.PoFile()
        msg = gettextpo.PoMessage()
        msg.set_msgid('Hello')
        poiter = iter(pofile)
        poiter.insert(msg)

        self.assertEquals(list(iter(pofile)), [msg])
    def testAddMessageTwice(self):
        '''A message object can only be added to one pofile object'''
        pofile1 = gettextpo.PoFile()
        pofile2 = gettextpo.PoFile()
        msg = gettextpo.PoMessage()
        msg.set_msgid('Hello')

        poiter = iter(pofile1)
        poiter.insert(msg)

        poiter = iter(pofile2)
        self.assertRaises(ValueError, poiter.insert, msg)

class PoMessageTestCase(unittest.TestCase):
    def testCreateMessage(self):
        '''Test that messages can be created.'''
        msg = gettextpo.PoMessage()

    def testSetMsgId(self):
        msg = gettextpo.PoMessage()
        msg.set_msgid('Hello')
        self.assertEquals(msg.msgid, 'Hello')
        msg.set_msgid_plural('Hellos')
        self.assertEquals(msg.msgid_plural, 'Hellos')

    def testSetMsgStr(self):
        msg = gettextpo.PoMessage()
        msg.set_msgstr('Hello World')
        self.assertEquals(msg.msgstr, 'Hello World')
    
    def testSetMsgStrPlural(self):
        '''Test handling of plural msgstrs.  The PoMessage object can
        not hold plural msgstrs if the msgid does not have a plural.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('Something')
        self.assertRaises(ValueError, msg.set_msgstr_plural, 0, 'Zero')
        self.assertEquals(msg.msgstr_plural, [])

        # need to set the plural msgid first, then add the plural msgstrs
        msg.set_msgid_plural('Somethings')
        msg.set_msgstr_plural(0, 'Zero')
        msg.set_msgstr_plural(1, 'One')
        msg.set_msgstr_plural(2, 'Two')
        self.assertEquals(msg.msgstr_plural, ['Zero', 'One', 'Two'])

class CheckFormatTestCase(unittest.TestCase):
    def testGoodFormat(self):
        '''Check that no exception is raised on a good translation.

        A translation may use different format characters while still
        being valid.  For example, both "%f" and "%g" can be used to
        format a floating point value, so no error should be raised on
        that kind of change.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('Hello %s %d %g')
        msg.set_format('c-format', True)
        msg.set_msgstr('Bye %s %.2d %f')

        # this should run without an exception
        msg.check_format()

    def testAddFormatSpec(self):
        '''Test that an exception is raised when a format string is added.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('No format specifiers')
        msg.set_format('c-format', True)
        msg.set_msgstr('One format specifier: %20s')
        self.assertRaises(gettextpo.error, msg.check_format)

    def testSwapFormatSpecs(self):
        '''Test that an exception is raised when format strings are transposed.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('Spec 1: %s, Spec 2: %d')
        msg.set_format('c-format', True)
        msg.set_msgstr('Spec 1: %d, Spec 2: %s')
        self.assertRaises(gettextpo.error, msg.check_format)

    def testNonFormatString(self):
        '''Test that no exception is raised if the message is not marked as a format string.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('Spec 1: %s, Spec 2: %d')
        msg.set_format('c-format', False)
        msg.set_msgstr('Spec 1: %d, Spec 2: %s')

        # this should run without an exception
        msg.check_format()

    def testEmptyMsgStr(self):
        '''Test that empty translations do not trigger a failure.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('Hello %s')
        msg.set_format('c-format', True)
        msg.set_msgstr(None)

        # this should run without an exception
        msg.check_format()

    def testGoodPlural(self):
        '''Test that a good plural message passes without error.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('%d apple')
        msg.set_msgid_plural('%d apples')
        msg.set_format('c-format', True)
        msg.set_msgstr_plural(0, '%d orange')
        msg.set_msgstr_plural(1, '%d oranges')
        msg.set_msgstr_plural(2, '%d oranges_')

        # this should run without an exception
        msg.check_format()

    def testBadPlural(self):
        '''Test that bad plural translations raise an error error.'''
        msg = gettextpo.PoMessage()
        msg.set_msgid('%d apple')
        msg.set_msgid_plural('%d apples')
        msg.set_format('c-format', True)
        msg.set_msgstr_plural(0, '%d orange')
        msg.set_msgstr_plural(1, '%d oranges')
        msg.set_msgstr_plural(2, '%g oranges_')
        self.assertRaises(gettextpo.error, msg.check_format)

    def testUnicodeString(self):
        """Test that a translation with unicode chars is working."""
        msg = gettextpo.PoMessage()
        msg.set_msgid(u'Carlos Perell\xf3 Mar\xedn')
        msg.set_msgstr(u'Carlos Perell\xf3 Mar\xedn')
        self.assertEqual(msg.msgid, 'Carlos Perell\xc3\xb3 Mar\xc3\xadn')
        self.assertEqual(msg.msgstr, 'Carlos Perell\xc3\xb3 Mar\xc3\xadn')

## XXXX - gettext doesn't seem to check for this one
#
#    def testBadPluralMsgId(self):
#        '''Test that conflicting plural msgids raise errors on their own.'''
#        msg = gettextpo.PoMessage()
#        msg.set_msgid('%d apple')
#        msg.set_msgid_plural('%g apples')
#        msg.set_format('c-format', True)
#        self.assertRaises(gettextpo.error, msg.check_format)
#

if __name__ == '__main__':
    unittest.main()
